﻿/*==============================================================================

   Name     : FMEServerRESTAPI.cpp

   System   : FME Server REST API

   Language : C++

   Purpose  : FMEServerAPIDemo sample application

   Author               Date            Changes made
   ------------------   ------------    -------------------------------
   Keith Huang          May 28, 2012    Original Definition
   Matheson Mawhinney   Nov 21, 2014    Updated to use REST V2 API

          Copyright (c) 1994 - 2014, Safe Software Inc.
                       All Rights Reserved

   The entire risk as to the results and performance of the software,
   supporting text and other information contained in this file
   (collectively called the "Software") is with the user.  Although
   Safe Software Incorporated has used considerable efforts in preparing
   the Software, Safe Software Incorporated does not warrant the
   accuracy or completeness of the Software. In no event will Safe Software
   Incorporated be liable for damages, including loss of profits or
   consequential damages, arising out of the use of the Software.

==============================================================================*/

#include <stdlib.h>
#include <iostream>
#include <string.h>
#include <QtNetwork/QtNetwork>
#include <QtCore/QtCore>

using namespace std;

class FMEServerRESTAPIDemo
{

public:

   FMEServerRESTAPIDemo();
   FMEServerRESTAPIDemo(string host, string port);
   void getConnectionInfo();
   void execute();
   void displayMainMenu();
   int getMenuChoice();
   void executeMenuChoice(int choice);
   void listJobs();
   void viewJob();
   void removeCompletedJobs();
   void cancelJob();
   void listRepositories();
   void addRepository();
   void removeRepository();
   void listWorkspaces();
   void addWorkspace();
   void updateWorkspace();
   void getWorkspace();
   void removeWorkspace();
   void listResources();
   void addResource();
   void updateResource();
   void getResource();
   void removeResource();
   void listServices();
   void addService();
   void updateService();
   void removeService();
   void runWorkspace();
   void listParameters();

private:

   string token_;
   string host_;
   string port_;
   string baseURL_;
   QNetworkAccessManager manager_;
};

//******************************HTTP METHODS*******************************

//========================================================================
//
QNetworkReply* postRequest(const QString &url, const QUrl &params, QNetworkAccessManager& manager)
{
   QNetworkRequest request;
   request.setUrl(url);
   request.setHeader(QNetworkRequest::ContentTypeHeader, "application/x-www-form-urlencoded");
   QByteArray data = params.toEncoded(); //add parameters

   QNetworkReply *reply = manager.post(request, data);
   QEventLoop loop;
   QObject::connect(reply, SIGNAL(finished()), &loop, SLOT(quit()));
   loop.exec();

   return reply;
}

//=========================================================================
//
QNetworkReply* postFile(const QString &url, const QByteArray &fileData, QNetworkAccessManager& manager, const string &workspaceName)
{
   QString dispositionHeader = QString::fromStdString("attachment; filename=\"" + workspaceName + "\"");
   QNetworkRequest request;
   request.setUrl(url);
   request.setHeader(QNetworkRequest::ContentDispositionHeader, QVariant(dispositionHeader));
   request.setHeader(QNetworkRequest::ContentTypeHeader, "application/octet-stream");

   QNetworkReply *reply = manager.post(request, fileData);
   QEventLoop loop;
   QObject::connect(reply, SIGNAL(finished()), &loop, SLOT(quit()));
   loop.exec();

   return reply;
}

//========================================================================
//
QNetworkReply* postRunWorkspace(const QString &url, const QByteArray &params, QNetworkAccessManager& manager)
{
   QNetworkRequest request;
   request.setUrl(url);
   request.setHeader(QNetworkRequest::ContentTypeHeader, "application/json");

   QNetworkReply *reply = manager.post(request, params);
   QEventLoop loop;
   QObject::connect(reply, SIGNAL(finished()), &loop, SLOT(quit()));
   loop.exec();

   return reply;
}

//=========================================================================
//
QNetworkReply* putRequest(const QString &url, const QUrl &fileData, QNetworkAccessManager& manager)
{
   QNetworkRequest request;
   request.setUrl(url);
   request.setHeader(QNetworkRequest::ContentTypeHeader, "application/x-www-form-urlencoded");
   QByteArray data = fileData.toEncoded();

   QNetworkReply *reply = manager.put(request, data);
   QEventLoop loop;
   QObject::connect(reply, SIGNAL(finished()), &loop, SLOT(quit()));
   loop.exec();

   return reply;
}

//========================================================================
//
QNetworkReply* putFile(const QString &url, const QByteArray &fileData, QNetworkAccessManager& manager, const string &workspaceName)
{
   QString dispositionHeader = QString::fromStdString("attachment; filename=\"" + workspaceName + "\"");
   QNetworkRequest request;
   request.setUrl(url);
   request.setHeader(QNetworkRequest::ContentDispositionHeader, QVariant(dispositionHeader));
   request.setHeader(QNetworkRequest::ContentTypeHeader, "application/octet-stream");

   QNetworkReply *reply = manager.put(request, fileData);
   QEventLoop loop;
   QObject::connect(reply, SIGNAL(finished()), &loop, SLOT(quit()));
   loop.exec();

   return reply;
}

//========================================================================
//
QNetworkReply* getDownload(const QString &url, QNetworkAccessManager& manager)
{
   QNetworkRequest request;
   request.setUrl(url);
   request.setRawHeader("Accept", "application/octet-stream");

   QNetworkReply *reply = manager.get(request);
   QEventLoop loop;
   QObject::connect(reply, SIGNAL(finished()), &loop, SLOT(quit()));
   loop.exec();

   return reply;
}

//=========================================================================
//
QNetworkReply* getRequest(const QString &url, QNetworkAccessManager& manager)
{
   QNetworkRequest request;
   request.setUrl(url);
   request.setRawHeader("Accept", "application/octet-stream");

   QNetworkReply *reply = manager.get(request);
   QEventLoop loop;
   QObject::connect(reply, SIGNAL(finished()), &loop, SLOT(quit()));
   loop.exec();

   return reply;
}

//=========================================================================
//
QNetworkReply* deleteRequest(const QString &url, QNetworkAccessManager& manager)
{
   QNetworkRequest request;
   request.setUrl(url);

   QNetworkReply *reply = manager.deleteResource(request);
   QEventLoop loop;
   QObject::connect(reply, SIGNAL(finished()), &loop, SLOT(quit()));
   loop.exec();

   return reply;
}

//****************************UTILITY FUNCTIONS****************************
string getInput(const string &promptMessage) // Output a prompt and get user input
{
   cout << promptMessage;
   char userInput[500];
   cin.getline(userInput, 500);
   string temp(userInput);
   return temp;
}

//*****************************CLASS FUNCTIONS*****************************

//=========================================================================
//
FMEServerRESTAPIDemo::FMEServerRESTAPIDemo() // Default constructor
{
   host_ = "localhost"; // Default host and port
   port_ = "80";
}

//=========================================================================
//
FMEServerRESTAPIDemo::FMEServerRESTAPIDemo(string host, string port) // Constructor
{
   host_ = host;
   port_ = port;
   baseURL_ = "http://" + host + ":" + port + "/fmerest/v2/";
   getConnectionInfo(); // Gets the token
}

//=========================================================================
//
void FMEServerRESTAPIDemo::getConnectionInfo()
{
   cout << "Please enter server connection credentials.\n";
   string userid_ = getInput("UserId:\n");

   if (userid_ != "") {
      string password_ = getInput("Password:\n");

      // We must first generate a security token for authentication purposes
      QString fmeUrl = QString::fromStdString("http://" + host_ + ":" + port_
         + "/fmetoken/service/generate?user=" + userid_ + "&password=" + password_ + "&generate=1&timeunit=hour");

      QNetworkReply *reply = getRequest(fmeUrl, manager_);

      // Custom error handling
      if (!reply->error())
      {
         QByteArray replyBytes = reply->readAll();
         string token(replyBytes.data());
         token_ = token;
      }
      else
      {
         QString qError = reply->errorString();
         string error = qError.toStdString();
         cout << error << "\n";
         return;
      }
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::execute() // Shows menu, gets input and performs action
{
   displayMainMenu();
   int choice = getMenuChoice(); // Has internal validation
   executeMenuChoice(choice);
}

//=========================================================================
//
void FMEServerRESTAPIDemo::displayMainMenu()
{
   cout <<
      "\n\n===FME Server REST API Demo===\n" <<
      "1. List available repositories\n" <<
      "2. Add a repository\n" <<
      "3. Remove a repository\n" <<
      "4. List available workspaces\n" <<
      "5. Add a workspace\n" <<
      "6. Update a workspace\n" <<
      "7. Get a workspace\n" <<
      "8. Remove a workspace\n" <<
      "9. List resources of a workspace\n" <<
      "10. Add a resource\n" <<
      "11. Update a resource\n" <<
      "12. Get a resource\n" <<
      "13. Remove a resource\n" <<
      "14. List available services\n" <<
      "15. Add a service\n" <<
      "16. Update a service\n" <<
      "17. Remove a service\n" <<
      "18. Run a workspace\n" <<
      "19. List published parameters of a workspace\n" <<
      "20. List jobs\n" <<
      "21. View a job\n" <<
      "0. Quit\n" <<
      "\nSELECT: ";
}

//=========================================================================
//
int FMEServerRESTAPIDemo::getMenuChoice() // Function has internal validation instead of throwing exceptions
{
   int choice;
   while (true){ // Validation
      cin >> choice;
      if (cin.fail()) // Input contained no numbers ie "five"
      {
         cin.clear();
         cin.ignore(1000, '\n');
         continue;
      }
      cin.ignore(1000, '\n'); // Clear garbage in input
      if (cin.gcount() > 1) // Input contained numbers and letters, so more than the new carriage character was cleared ie "5five"
      {
         cout << "Invalid input\n";
         continue;
      }
      if (choice < 0 || choice > 21)// Input contained invalid choice
      {
         cout << "Invalid input\n";
         continue;
      }
      break;
   }

   return choice;
}

//=========================================================================
//
void FMEServerRESTAPIDemo::executeMenuChoice(int choice)
{
   switch (choice)
   {
   case 1:
      listRepositories();
      break;
   case 2:
      addRepository();
      break;
   case 3:
      removeRepository();
      break;
   case 4:
      listWorkspaces();
      break;
   case 5:
      addWorkspace();
      break;
   case 6:
      updateWorkspace();
      break;
   case 7:
      getWorkspace();
      break;
   case 8:
      removeWorkspace();
      break;
   case 9:
      listResources();
      break;
   case 10:
      addResource();
      break;
   case 11:
      updateResource();
      break;
   case 12:
      getResource();
      break;
   case 13:
      removeResource();
      break;
   case 14:
      listServices();
      break;
   case 15:
      addService();
      break;
   case 16:
      updateService();
      break;
   case 17:
      removeService();
      break;
   case 18:
      runWorkspace();
      break;
   case 19:
      listParameters();
      break;
   case 20:
      listJobs();
      break;
   case 21:
      viewJob();
      break;
   case 0:
      exit(-1);
   default:
      cout << "Invalid option!";
      return;
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::listRepositories()
{
   QString fmeUrl = QString::fromStdString(baseURL_ + "repositories?token=" + token_);

   QNetworkReply *reply = getRequest(fmeUrl, manager_);

   if (!reply->error())
   {
      QByteArray strReply = reply->readAll();
      QJsonDocument document = QJsonDocument::fromJson(strReply);
      QJsonArray array = document.array();

      cout << "=============================================\n";
      for (const QJsonValue & value : array)
      {
         QJsonObject obj = value.toObject();
         cout << "Name: " << obj.value("name").toString().toStdString() << "\n";
         cout << "Description: " << obj.value("description").toString().toStdString() << "\n";
         cout << "=============================================\n";
      }
   }
   else
   {
      QString qError = reply->errorString();
      string error = qError.toStdString();
      cout << error << "\n";
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::addRepository()
{
   string name = getInput("Please enter the name for the new repository:\n");
   string description = getInput("Please enter the description for the new repository:\n");

   QString fmeUrl = QString::fromStdString(baseURL_ + "repositories?token=" + token_);

   QUrl params;
   QUrlQuery query;
   query.addQueryItem("token", QString::fromStdString(token_));
   query.addQueryItem("name", QString::fromStdString(name));
   query.addQueryItem("description", QString::fromStdString(description));
   params.setQuery(query);
   QNetworkReply *reply = postRequest(fmeUrl, params, manager_);

   if (reply->error())
   {
      QString qError = reply->errorString();
      string error = qError.toStdString();
      cout << error << "\n" << "Repository creation failed";
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::removeRepository()
{
   string name = getInput("Please enter the name of the repository to be removed:\n");

   QString fmeUrl = QString::fromStdString(baseURL_ + "repositories/" + name + "?token=" + token_);

   QNetworkReply *reply = deleteRequest(fmeUrl, manager_);

   if (reply->error())
   {
      QString qError = reply->errorString();
      string error = qError.toStdString();
      cout << error << "\n" << "Repository deletion failed";
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::listWorkspaces()
{
   string name = getInput("Please enter the name of the repository to show workspaces for:\n");

   // We must first generate a security token for authentication purposes
   QString fmeUrl = QString::fromStdString(baseURL_ + "repositories/" + name + "/items?token=" + token_);

   QNetworkReply *reply = getRequest(fmeUrl, manager_);

   if (!reply->error())
   {
      QByteArray replyStr = reply->readAll();
      QJsonDocument document = QJsonDocument::fromJson(replyStr);
      QJsonArray array = document.array();

      cout << "=============================================\n";
      for (const QJsonValue & value : array)
      {
         QJsonObject obj = value.toObject();
         cout << "Name: " << obj.value("name").toString().toStdString() << "\n";
         cout << "=============================================\n";
      }
   }
   else
   {
      QString qError = reply->errorString();
      string error = qError.toStdString();
      cout << error << "\n";
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::addWorkspace()
{
   string repositoryName = getInput("Plese enter the repository name for the new workspace:\n");
   string workspaceFilePath = getInput("Please enter the file path of the new workspace to add:\n");

   QFile file = (QString::fromStdString(workspaceFilePath));

   if (!file.open(QIODevice::ReadOnly))
   {
      cout << "File cannot be opened or does not exist";
      return;
   }
   else
   {
      QByteArray data(file.readAll());
      file.close();

      QFileInfo fileInfo(file.fileName());
      QString filename(fileInfo.fileName()); // Retrieve filename
      string workspaceName(filename.toStdString());
      QString fmeUrl = QString::fromStdString(baseURL_ + "repositories/" + repositoryName + "/items?token=" + token_);

      QNetworkReply *reply = postFile(fmeUrl, data, manager_, workspaceName);
      if (reply->error())
      {
         QString qError = reply->errorString();
         string error = qError.toStdString();
         cout << error << "\n" << "Adding workspace failed";
      }
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::updateWorkspace()
{
   string repositoryName = getInput("Please enter the name of the repository containing the workspace to update:\n");
   string workspaceFilePath = getInput("Please enter the local file path of the new workspace:\n");
   string workspaceToReplace = getInput("Please enter the name of the workspace to update:\n");

   QFile file = (QString::fromStdString((workspaceFilePath)));

   if (!file.open(QIODevice::ReadOnly))
   {
      cout << "File cannot be opened or does not exist";
      return;
   }
   else
   {
      QByteArray data(file.readAll());
      file.close();

      QFileInfo fileInfo(file.fileName());
      QString filename(fileInfo.fileName());
      string workspaceName(filename.toStdString());
      QString fmeUrl = QString::fromStdString(baseURL_ + "repositories/" + repositoryName + "/items/" + workspaceToReplace +
         "?token=" + token_);

      QNetworkReply *reply = putFile(fmeUrl, data, manager_, workspaceName);
      if (reply->error())
      {
         QString qError = reply->errorString();
         string error = qError.toStdString();
         cout << error << "\n" << "Adding workspace failed";
      }
   }
}


//=========================================================================
//
void FMEServerRESTAPIDemo::getWorkspace()
{
   string repositoryName = getInput("Please enter the repository name for the workspace to get:\n");
   string workspaceName = getInput("Please enter the workspace name (with extensions):\n");
   string workspaceFilePath = getInput("Please enter the local file path for the workspace to be stored:\n");

   QString fmeUrl = QString::fromStdString(baseURL_ + "repositories/" + repositoryName + "/items/" + workspaceName + "?token=" + token_);

   QNetworkReply *reply = getDownload(fmeUrl, manager_);

   if (!reply->error())
   {
      QByteArray replyeBytes = reply->readAll();
      QFile file(QString::fromStdString(workspaceFilePath));
      file.open(QIODevice::ReadWrite);
      QTextStream out(&file);
      out << replyeBytes.data();
      file.close();
   }
   else
   {
      QString qError = reply->errorString();
      string error = qError.toStdString();
      cout << error << "\n";
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::removeWorkspace()
{
   string repositoryName = getInput("Please enter the repository name for the workspace to remove:\n");
   string workspaceName = getInput("Please enter the name for the workspace to remove:\n");

   QString fmeUrl = QString::fromStdString(baseURL_ + "repositories/" + repositoryName + "/items/" + workspaceName + "?token=" + token_);

   QNetworkReply *reply = deleteRequest(fmeUrl, manager_);

   if (reply->error())
   {
      QString qError = reply->errorString();
      string error = qError.toStdString();
      cout << error << "\n" << "Repository deletion failed";
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::listResources()
{
   string repositoryName = getInput("Please enter the repository name for the resources:\n");
   string workspaceName = getInput("Please enter the workspace name for the resources:\n");

   // We must first generate a security token for authentication purposes
   QString fmeUrl = QString::fromStdString(baseURL_ + "repositories/" + repositoryName + "/items/" + workspaceName + "/resources?token=" + token_);

   QNetworkReply *reply = getRequest(fmeUrl, manager_);

   if (!reply->error())
   {
      QByteArray replyStr = reply->readAll();
      QJsonDocument document = QJsonDocument::fromJson(replyStr);
      QJsonArray array = document.array();

      cout << "=============================================\n";
      for (const QJsonValue & value : array)
      {
         QJsonObject obj = value.toObject();
         cout << "Name: " << obj.value("name").toString().toStdString() << "\n";
         cout << "=============================================\n";
      }
   }
   else
   {
      QString qError = reply->errorString();
      string error = qError.toStdString();
      cout << error << "\n";
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::addResource()
{
   string repositoryName = getInput("Please enter the repository name for the new resource:\n");
   string workspaceName = getInput("Please enter the workspace name for the new resource:\n");
   string resourceFilePath = getInput("Please enter the file path of the new resource to add (with extensions):\n");

   QFile file(QString::fromStdString(resourceFilePath));
   QFileInfo fileInfo(file.fileName());
   QString filename(fileInfo.fileName().section(".", 0, 0));
   string resourceName = filename.toStdString();

   if (!file.open(QIODevice::ReadOnly))
   {
      cout << "File cannnot be opened or does not exist";
      getchar();
      return;
   }
   else
   {
      QByteArray data(file.readAll());
      file.close();

      QString fmeUrl = QString::fromStdString(baseURL_ + "repositories/" + repositoryName + "/items/" + workspaceName
         + "/resources?token=" + token_);

      QNetworkReply *reply = postFile(fmeUrl, data, manager_, resourceName);

      if (reply->error())
      {
         QString qError = reply->errorString();
         string error = qError.toStdString();
         cout << error << "\n" << "Adding resource failed";
      }
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::updateResource()
{
   string repositoryName = getInput("Please enter the repository name for the new resource:\n");
   string workspaceName = getInput("Please enter the workspace name for the new resource:\n");
   string existingResource = getInput("Please enter the name of the resource to be replaced:\n");
   string resourceFilePath = getInput("Please enter the file path of the new resource to add (with extensions):\n");

   QFile file(QString::fromStdString(resourceFilePath));
   QFileInfo fileInfo(file.fileName());
   QString filename(fileInfo.fileName());
   string resourceName = filename.toStdString();

   if (!file.open(QIODevice::ReadOnly))
   {
      cout << "File cannnot be opened or does not exist";
      getchar();
      return;
   }
   else
   {
      QByteArray data(file.readAll());
      file.close();

      QString fmeUrl = QString::fromStdString(baseURL_ + "repositories/" + repositoryName + "/items/" + workspaceName
         + "/resources/" + existingResource + "?token=" + token_);

      QNetworkReply *reply = putFile(fmeUrl, data, manager_, resourceName);

      if (reply->error())
      {
         QString qError = reply->errorString();
         string error = qError.toStdString();
         cout << error << "\n" << "Adding resource failed";
      }
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::getResource()
{
   string repositoryName = getInput("Please enter the repository name for the resource to get:\n");
   string workspaceName = getInput("Please enter the workspace name for the resource to get:\n");
   string resourceName = getInput("Please enter the resource name (with extensions) to get:\n");
   string resourceFilePath = getInput("Please enter the local file path for the resource to be stored:\n");

   QString fmeUrl = QString::fromStdString(baseURL_ + "repositories/" + repositoryName + "/items/" + workspaceName +
      "/resources/" + resourceName + "?token=" + token_);

   QNetworkReply *reply = getRequest(fmeUrl, manager_);

   if (!reply->error())
   {
      QByteArray replyeBytes = reply->readAll();
      QFile file(QString::fromStdString(resourceFilePath));
      file.open(QIODevice::ReadWrite);
      QTextStream out(&file);
      out << replyeBytes.data();
      file.close();
   }
   else
   {
      QString qError = reply->errorString();
      string error = qError.toStdString();
      cout << error << "\n";
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::removeResource()
{
   string repositoryName = getInput("Please enter the repository name for the resource to be removed:\n");
   string workspaceName = getInput("Please enter the workspace name for the resource to be removed:\n");
   string resourceName = getInput("Please enter the resource name (with extensions) to be removed:\n");

   QString fmeUrl = QString::fromStdString(baseURL_ + "repositories/" + repositoryName + "/items/" + workspaceName +
      "/resources/" + resourceName + "?token=" + token_);

   QNetworkReply *reply = deleteRequest(fmeUrl, manager_);

   if (reply->error())
   {
      QString qError = reply->errorString();
      string error = qError.toStdString();
      cout << error << "\n" << "Repository deletion failed";
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::listServices()
{
   QString fmeUrl = QString::fromStdString(baseURL_ + "services?token=" + token_);

   QNetworkReply *reply = getRequest(fmeUrl, manager_);

   if (!reply->error())
   {
      QByteArray replyStr = reply->readAll();
      QJsonDocument document = QJsonDocument::fromJson(replyStr);
      QJsonArray array = document.array();

      cout << "=============================================\n";
      for (const QJsonValue & value : array)
      {
         QJsonObject obj = value.toObject();
         cout << "Name: " << obj.value("name").toString().toStdString() + "\n";
         cout << "=============================================\n";
      }
   }
   else
   {
      QString qError = reply->errorString();
      string error = qError.toStdString();
      cout << error << "\n";
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::addService()
{
   string name = getInput("Please enter the name of the new service:\n");
   string urlPattern = getInput("Please enter the URL pattern of the new service:\n");
   string displayName = getInput("Please enter the display name of the new service:\n");
   string description = getInput("Please enter the description of the new service:\n");

   QString fmeUrl = QString::fromStdString(baseURL_ + "services?token=" + token_);

   QUrl url;
   QUrlQuery query;
   query.addQueryItem("token", QString::fromStdString(token_));
   query.addQueryItem("name", QString::fromStdString(name));
   query.addQueryItem("description", QString::fromStdString(description));
   query.addQueryItem("displayName", QString::fromStdString(displayName));
   query.addQueryItem("url", QString::fromStdString(urlPattern));
   url.setQuery(query);

   QNetworkAccessManager *manager = new QNetworkAccessManager();
   QNetworkReply *reply = postRequest(fmeUrl, url, manager_);

   if (reply->error())
   {
      QString qError = reply->errorString();
      string error = qError.toStdString();
      cout << error << "\n" << "Service addition failed";
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::updateService()
{
   string oldName = getInput("Please enter the name of the existing service to update:\n");
   string newName = getInput("Please enter the name of the new service:\n");
   string urlPattern = getInput("Please enter the URL pattern of the new service:\n");
   string displayName = getInput("Please enter the display name of the new service:\n");
   string description = getInput("Please enter the description of the new service:\n");

   QString fmeUrl = QString::fromStdString(baseURL_ + "services/" + oldName + "?token=" + token_);

   QUrl url;
   QUrlQuery query;
   query.addQueryItem("token", QString::fromStdString(token_));
   query.addQueryItem("name", QString::fromStdString(newName));
   query.addQueryItem("description", QString::fromStdString(description));
   query.addQueryItem("displayName", QString::fromStdString(displayName));
   query.addQueryItem("url", QString::fromStdString(urlPattern));
   url.setQuery(query);

   QNetworkReply *reply = putRequest(fmeUrl, url, manager_);

   if (reply->error())
   {
      QString qError = reply->errorString();
      string error = qError.toStdString();
      cout << error << "\n" << "Service addition failed";
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::removeService()
{
   string serviceName = getInput("Please enter name of the service to be remove:\n");

   QString fmeUrl = QString::fromStdString(baseURL_ + "services/" + serviceName + "?token=" + token_);

   QNetworkReply *reply = deleteRequest(fmeUrl, manager_);

   if (reply->error())
   {
      QString qError = reply->errorString();
      string error = qError.toStdString();
      cout << error << "\n" << "Repository deletion failed";
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::runWorkspace()
{
   string repositoryName = getInput("Please enter the name for the repository:\n");
   string workspaceName = getInput("Please enter the name for the workspace to run:\n");

   QByteArray body = "{}";
   QString fmeUrl = QString::fromStdString(baseURL_ + "transformations/commands/transact/" + repositoryName +
      "/" + workspaceName + "?token=" + token_);

   QNetworkReply *reply = postRunWorkspace(fmeUrl, body, manager_);

   if (!reply->error())
   {
      QByteArray replyStr = reply->readAll();
      QJsonDocument document = QJsonDocument::fromJson(replyStr);
      QJsonObject json = document.object();

      cout << "=============================================\n";
      cout << "ID: " << json.value("id").toInt() << "\n";
      cout << "Status: " << json.value("status").toString().toStdString() << "\n";
      cout << "Status Message: " << json.value("statusMessage").toString().toStdString() << "\n";
      cout << "timeFinished: " << json.value("timeFinished").toString().toStdString() << "\n";
      cout << "=============================================\n";
   }
   else
   {
      QString qError = reply->errorString();
      string error = qError.toStdString();
      cout << error << "\n";
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::listParameters()
{
   string repositoryName = getInput("Please enter the repository name for the workspace:\n");
   string workspaceName = getInput("Please enter the name of the workspace to view the parameters of:\n");

   QString fmeUrl = QString::fromStdString(baseURL_ + "repositories/" + repositoryName + "/items/" +
      workspaceName + "/parameters?token=" + token_);

   QNetworkReply *reply = getRequest(fmeUrl, manager_);

   if (!reply->error())
   {
      QByteArray replyStr = reply->readAll();
      QJsonDocument document = QJsonDocument::fromJson(replyStr);
      QJsonArray array = document.array();

      cout << "=============================================\n";
      for (const QJsonValue & value : array)
      {
         QJsonObject obj = value.toObject();
         cout << "Name: " << obj.value("name").toString().toStdString() << "\n";
         cout << "Description: " << obj.value("description").toString().toStdString() << "\n";
         cout << "=============================================\n";
      }
   }
   else
   {
      QString qError = reply->errorString();
      string error = qError.toStdString();
      cout << error << "\n";
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::listJobs()
{
   // We would like to see the IDs of all jobs listed
   // So, we perform the same action for each job type
   string categories[] = {
      "running",
      "completed",
      "queued"
   };

   for (int i = 0; i < 3; i++)
   {
      QString fmeUrl = QString::fromStdString(baseURL_ + "transformations/jobs/" + categories[i] + "?token=" + token_);

      QNetworkReply *reply = getRequest(fmeUrl, manager_);

      cout << "\n" << categories[i] << ":\n";

      if (!reply->error())
      {
         QByteArray replyStr = reply->readAll();
         QJsonDocument document = QJsonDocument::fromJson(replyStr);
         QJsonArray array = document.array();

         cout << "=============================================\n";
         for (const QJsonValue &value : array)
         {
            QJsonObject obj = value.toObject();
            cout << "ID: " << obj.value("id").toInt() << "\n";
            cout << "Status: " << obj.value("status").toString().toStdString() << "\n";
            cout << "Time Finished: " << obj.value("timeFinished").toString().toStdString() << "\n";
            cout << "=============================================\n";
         }
      }
      else
      {
         QString qError = reply->errorString();
         string error = qError.toStdString();
         cout << error << "\n";
      }
   }
}

//=========================================================================
//
void FMEServerRESTAPIDemo::viewJob()
{
   string id = getInput("Please enter the id of the job to view:\n");
   QString fmeUrl = QString::fromStdString(baseURL_ + "transformations/jobs/id/" + id + "?token=" + token_);

   QNetworkReply *reply = getRequest(fmeUrl, manager_);

   if (!reply->error()) // If no error, will return NoError (which has enum = 0)
   {
      QByteArray replyStr = reply->readAll();
      QJsonDocument document = QJsonDocument::fromJson(replyStr);
      QJsonObject obj = document.object();

      cout << "=============================================\n";
      cout << "ID: " << obj.value("id").toInt() << "\n";
      cout << "Status: " << obj.value("status").toString().toStdString() << "\n";
      cout << "Time Finished: " << obj.value("timeFinished").toString().toStdString() << "\n";
      cout << "=============================================\n";
   }
   else
   {
      QString qError = reply->errorString();
      string error = qError.toStdString();
      cout << error << "\n";
   }
}

//=========================================================================
//
void argError()
{
   cout << "USAGE:\n\tFMEServerRESTAPIDemo <host> <port>";
   getchar();
   getchar();
   exit(-1);
}

int main(int argc, char *argv[])
{
   QCoreApplication app(argc, argv);

   if (argc != 3)
   {
      argError();
   }
   else
   {
      string host(argv[1]), port(argv[2]);
      FMEServerRESTAPIDemo demo(host, port);
      while (true)
      {
         demo.execute();
      }
   }
   getchar();
   getchar();

   return 0;
}
